#!/usr/bin/python
import sys,os, struct
import string
from structures import pe_table, pv_table, lv_table, vg_table, uuidlist_table
from types import disk_struct


consistent=1

def inconsistency(type,instance):
	global consistent
	consistent=0
	if permissive:
		print "ERROR: %s (%s)"%(type,instance)
	else:
		raise type,instance

def printornot(what):
	global quiet
	if not quiet:
		print what

def okmessage(what):
	global consistent
	if consistent:
		printornot("%s : OK"%what)
	else:
		consistent=1
		print "%s : INCONSISTENT"%what

def check(struct,field,check):
	if not eval("%s %s"%(struct.get_val(field),check)):
		inconsistency(field,"is not %s"%(check))
	else:
		okmessage(field)

def check2(struct1,field1,check,struct2,field2):
	if not eval("%s %s %s "%(
		struct1.get_val(field1),
		check,
		struct2.get_val(field2))):
		inconsistency(field1,"%s %s %s is not true"%(field1,check,field2))
	else:
		okmessage(field1)

def check3(struct1,field1,check1,struct2,field2,check2,struct3,field3):
	if not eval("%s %s %s %s %s"%(
		struct1.get_val(field1),
		check1,
		struct2.get_val(field2),
		check2,
		struct3.get_val(field3),
		)):
		inconsistency(field1,"%s %s %s %s %s is not true"%(field1,check1,field2,check2,field3))
	else:
		okmessage(field1)


class pvcheck:
	def __init__(self,partition):
		pv=os.open(partition,os.O_RDONLY)
		pv_struct=disk_struct(pv,0,pv_table)
		self.pv_struct=pv_struct
		if verbose:
			print("***********************************PV***********************")
			pv_struct.display()
			print("***********************************End of PV***********************")
		
		vgbase=pv_struct.get_val("vg_on_disk_base")
		vg_struct=disk_struct(pv,int(vgbase),vg_table)
		self.vg_struct=vg_struct
		if verbose:
			print("***********************************VG***********************")
			vg_struct.display()
			print("***********************************End of VG***********************")
		
		
		
		check(pv_struct,"version"," == 1")
		check(pv_struct,"pv_on_disk_base"," == 0")
		check(pv_struct,"pv_on_disk_size"," == 1024")
		check(pv_struct,"vg_on_disk_base"," == 4096")
		check(pv_struct,"vg_on_disk_size"," == 4096")
		check(pv_struct,"pv_uuidlist_on_disk_base"," == 8192")
		check2(pv_struct,"pv_uuidlist_on_disk_size",
			" == 128 + 128 * ",
			vg_struct,"pv_max")
		check3(pv_struct,"lv_on_disk_base",
			" == ",
			pv_struct,"pv_uuidlist_on_disk_base",
			" + ", 
			pv_struct,"pv_uuidlist_on_disk_size")
		check2(pv_struct,"lv_on_disk_size",
			" == 328 + 328 *",
			vg_struct,"lv_max")
		check3(pv_struct,"pe_on_disk_base",
			" >= ",
			pv_struct,"lv_on_disk_base",
			" + ", 
			pv_struct,"lv_on_disk_size")
		check2(pv_struct,"pe_on_disk_size",
			" >= 4 + 4 * ",
			pv_struct,"pe_total")
		
		check3(pv_struct, "pv_size",
			" >= ",
			pv_struct, "pe_size",
			" * ",
			pv_struct, "pe_total")
		
		check2(pv_struct, "lv_cur",
			" <= ",
			vg_struct, "lv_cur")
		check2(pv_struct, "pe_total",
			" >= ",
			pv_struct, "pe_allocated")
		
		check(pv_struct, "pv_allocatable",
			" & 0xfffffffd == 0")
		check(pv_struct, "pv_status",
			" & 0xfffffffe == 0")
		
		#checks of vg
		check(vg_struct, "vg_status",
			" & 0xfffffff8 == 0")
		check(vg_struct, "vg_access",
			" & 0xfffffff0 == 0")
		check2(vg_struct, "lv_max",
			" >= ",
			vg_struct, "lv_cur")
		check2(vg_struct, "lv_cur",
			" >= ",
			vg_struct, "lv_open")
		check2(vg_struct, "pv_max",
			" >= ",
			vg_struct, "pv_cur")
		check2(vg_struct, "pv_cur",
			" >= ",
			vg_struct, "pv_act")
		check2(vg_struct, "pe_size",
			" == ",
			pv_struct, "pe_size")
		check2(pv_struct, "pe_total",
			" >= ",
			pv_struct, "pe_allocated")
		
		
		
		lvtab_base=pv_struct.get_val("lv_on_disk_base")
		lv_max=vg_struct.get_val("lv_max")
		
		self.lvtab=[]
		for off in range(0,lv_max):
			lvstruct=disk_struct(pv,lvtab_base+off*328,lv_table)
		
		
			self.lvtab.append(lvstruct)
			lv_name=lvstruct.get_val("lv_name")
			if verbose:
				print("***********************************LV # %u ***********************"%off)
				if not veryverbose:
					if(lv_name[0] == '\0'):
						print "lv name is empty"
					else:
						lvstruct.display()
				else:
					lvstruct.display()
				print("***********************************END of LV # %u ***********************"%off)

		self.uuidlist=[]
		uuidlist_base=pv_struct.get_val("pv_uuidlist_on_disk_base")
		if verbose:
			print("***********************************UUIDLIST ***********************")
		for uuidnum in range(0,vg_struct.get_val("pv_cur")):
			uuid_struct=disk_struct(pv,uuidlist_base+128*uuidnum,uuidlist_table)
			self.uuidlist.append(uuid_struct)
			if verbose:
				uuid_struct.display()
		if verbose:
			print("***********************************UUIDLIST END***********************")

		self.pelist=[]
		self.lv_pe={}
		pelist_base=pv_struct.get_val("pe_on_disk_base")
		pv_number=pv_struct.get_val("pv_number")
		if verbose:
			print("***********************************PE LIST ***********************")
		for penum in range(0,pv_struct.get_val("pe_total")):
			pe_struct=disk_struct(pv,pelist_base+4*penum,pe_table)
			self.pelist.append(pe_struct)
			lv=pe_struct.get_val("lv_num")
			le=pe_struct.get_val("le_num")
			if lv:
				if not self.lv_pe.has_key(lv):
					self.lv_pe[lv]=[]
				self.lv_pe[lv].append([le,pv_number,penum])
			if verbose:
				print "pe# =%u"%penum
				pe_struct.display()
		if verbose:
			print("*********************************** PE LIST END***********************")
		if len(self.lv_pe.keys()) != pv_struct.get_val("lv_cur"):
			#FIXME
			inconsistency("PE allocation table has different number of LVs than PV struct","lv list: %s, lv_cur=%u"%(self.lv_pe.keys(),pv_struct.get_val("lv_cur")))
			okmessage("PE allocation table preliminary test")

args= sys.argv[1:]

helptext=""
def mygetopt(opt,name):
	global args
	global helptext
	helptext=helptext+"%s\t:\t%s\n"%(opt,name)
	hasit=0
	try:
		hasit=args.index(opt)
		args.remove(opt)
		print "IS %s"%name
		hasit=1
	except:
		print "NO %s"%name
	return hasit
		
verbose=mygetopt("-v","verbose reporting")
veryverbose=mygetopt("-w","very verbose reporting")
permissive=mygetopt("-p","permissive execution (no halt on errors)")
quiet=mygetopt("-q","quiet execution")
help=mygetopt("-h","help (and nothing is done)")

if help:
	print helptext
	sys.exit()

pvlist=[]
pvhash={}
#pv1=pvcheck("pvdata.part3")

for pvname in args:
	printornot("***checking %s"%pvname)
	pv=pvcheck(pvname)
	pvlist.append(pv)
	pv_num=pv.pv_struct.get_val("pv_number")
	pvhash[pv_num]=pv

masterpv=pvlist[0]
if len(args) != masterpv.vg_struct.get_val("pv_cur"):
	inconsistency("number of physical volumes is not the same as in the command line","pv_cur=%s, args=%s"%(masterpv.vg_struct.get_val("pv_cur"),args))
for pv in pvlist[1:]:
	pv_number=pv.pv_struct.get_val("pv_number")
	printornot("***checking pv copy consistency for pv #%u"%pv_number)
	for field in ["id","version","vg_name","system_id"]:
		if pv.pv_struct.get_val(field) != masterpv.pv_struct.get_val(field):
			inconsistency("field %s in pv copy is different"%field,"%s != %s"%(pv.pv_struct.get_val(field),masterpv.pv_struct.get_val(field)))
	okmessage("pv copy consistency for pv #%u"%pv_number)

	printornot("***checking vg copy consistency for pv #%u"%pv_number)
	for row in vg_table:
		field=row[0]
		if pv.vg_struct.get_val(field) != masterpv.vg_struct.get_val(field):
			inconsistency("field %s in vg copy is different"%field,"%s != %s"%(pv.vg_struct.get_val(field) != masterpv.vg_struct.get_val(field)))
	okmessage("vg copy consistency for pv #%u"%pv_number)
	printornot("***checking lv copy consistency for pv #%u"%pv_number)
	for lv_num in range(0,masterpv.vg_struct.get_val("lv_max")):
		#FIXME: is there any upper meaning of the last word of my last lv entry is being 0xffff?
		lv_name=pv.lvtab[lv_num].get_val("lv_name")
		lv_name_master=masterpv.lvtab[lv_num].get_val("lv_name")
		if((lv_name[0] == '\0') and (lv_name_master[0] == '\0')):
			continue
		for row in lv_table:
			field=row[0]
			if pv.lvtab[lv_num].get_val(field) != masterpv.lvtab[lv_num].get_val(field):
				inconsistency("field %s is different in lv copy"%(field),"in lv #%u %s != %s"%(lv_num,pv.lvtab[lv_num].get_val(field),masterpv.lvtab[lv_num].get_val(field)))
	okmessage("lv copy consistency for pv #%u"%pv_number)
	printornot("***checking pv_uuid copy consistency for pv #%u"%pv_number)
	for pv_num in range(0,masterpv.vg_struct.get_val("pv_cur")):
		for row in uuidlist_table:
			field=row[0]
			if pv.uuidlist[pv_num].get_val(field) != masterpv.uuidlist[pv_num].get_val(field):
				inconsistency("field %s is different in pv_uuidlist copy"%(field),"for pv #%u %s != %s"%(pv_num,pv.uuidlist[pv_num].get_val(field),masterpv.uuidlist[pv_num].get_val(field)))
	okmessage("pv_uuid copy consistency for pv #%u"%pv_number)


printornot("***checking pv_uuid consistency")
for pv_num in range(0,masterpv.vg_struct.get_val("pv_cur")):
	pv=pvhash[pv_num+1]
	uuid_from_pv=pv.pv_struct.get_val("pv_uuid")
	uuid_from_uuidlist=pv.uuidlist[pv_num].get_val("pv_uuid")
	if(uuid_from_pv != uuid_from_uuidlist):
		inconsistency("pv_uuidlist wrong",
			"from pv      :%s\nfrom uuidlist:%s"%(
				uuid_from_pv, uuid_from_uuidlist))
okmessage("pv_uuid consistency")

printornot("***checking pe allocation table consistency")
le_sum={}
for pv in pvlist:
	for lv in pv.lv_pe.keys():
		if not le_sum.has_key(lv):
			le_sum[lv]={}
		for row in pv.lv_pe[lv]:
			(le,pv_num,pe_num)=row
			if le_sum[lv].has_key(le):
				inconsistency("duplicate logical extent",
					"lv=%u,(pv=%u,pe=%u),(pv=%u,pe=%u)"%(
						lv,pv_num,pe_num,
						le_sum[lv][le][0],le_sum[lv][le][1]))
			le_sum[lv][le]=[pv_num,pe_num]
pv=pvlist[0]
pe_size=pv.vg_struct.get_val("pe_size")
id=string.maketrans("a","a")
for lv in le_sum.keys():
	lv_len=pv.lvtab[lv-1].get_val("lv_size")/pe_size
	lv_name=string.translate(pv.lvtab[lv-1].get_val("lv_name"),id,'\0')
#	if lv_name[0]=='\0':
#		continue
	printornot("***checking lv %s(%u)"%(lv_name,lv))
	letab=le_sum[lv].keys()
	letab.sort()
	if(letab!=range(0,lv_len)):
		inconsistency("pe allocation is wrong","lv=%s(%u),lv_len=%u,lelist=%s"%(
			lv_name,lv,lv_len,letab))
okmessage("pe allocation table consistency")

